# -*- coding: utf-8 -*-

import logging
import math
import os
import os.path
import resource
import subprocess
import sys

from . import portfolio_runner


DRIVER_DIR = os.path.abspath(os.path.dirname(__file__))
SRC_DIR = os.path.dirname(DRIVER_DIR)
TRANSLATE = os.path.join(SRC_DIR, "translate", "translate.py")
PREPROCESS = os.path.join(SRC_DIR, "preprocess", "preprocess")
SEARCH_DIR = os.path.join(SRC_DIR, "search")


def call_cmd(cmd, args, stdin=None, preexec_funct=None):
    sys.stdout.flush()
    if stdin:
        with open(stdin) as stdin_file:
            if (preexec_funct):
                subprocess.check_call([cmd] + args, stdin=stdin_file, preexec_fn=preexec_funct)
            else:
                subprocess.check_call([cmd] + args, stdin=stdin_file)
                
    else:
        if (preexec_funct):
            subprocess.check_call([cmd] + args, preexec_fn=preexec_funct)
        else:
            subprocess.check_call([cmd] + args)

def set_number_black_variables(args, number_black):
    # TODO for now, assume that the redblack option is of form
    # option(some_parameters), where some_parameters may contain
    # number_black_vars, which in this case will be overwritten
    # TODO probably also remove the ratio_black option if present
    for i in range(0, len(args.search_options)):
        arg = args.search_options[i]
        if (arg == "--redblack"):
            option = args.search_options[i+1]
            
            # TODO there should be some nicer way to do all this
            if ("incremental_search=false, " in option):
                option = option.replace("incremental_search=false, ", "")
            if ("incremental_search=false," in option):
                option = option.replace("incremental_search=false,", "")
            if ("incremental_search=false" in option):
                option = option.replace("incremental_search=false", "")
            
            br_open = option.find("(")
            br_close = option.find(")")
            
            if (br_open == -1 or br_close == -1):
                print("redblack option should look like this: \"option(parameters)\", but it looks like this:")
                print(option)
                exit(1)
            
            num_vars_string = "number_black_vars="
            inc_flag_string = "incremental_search=true"
            if (num_vars_string in option):
                ind = option.find(num_vars_string)                
                new_option = option[0:ind + len(num_vars_string)]
                new_option += str(number_black)
                if (not inc_flag_string in option):
                    new_option += ", " + inc_flag_string
                new_option += option[min(option.find(",", ind), br_close):]
                args.search_options[i+1] = new_option
                print("changed redblack option to: " + new_option)
            else:
                new_option = option[0:br_open+1]
                new_option += num_vars_string
                new_option += str(number_black)
                if (not inc_flag_string in option):
                    new_option += ", " + inc_flag_string
                if (br_close - br_open != 1):
                    new_option += ", "
                    new_option += option[br_open+1:]
                else:
                    new_option += option[br_close:]
                args.search_options[i+1] = new_option
                print("changed redblack option to: " + new_option)
            break
    

def run_translate(args):
    logging.info("Running translator.")
    logging.info("translator inputs: %s" % args.translate_inputs)
    logging.info("translator arguments: %s" % args.translate_options)
    call_cmd(TRANSLATE, args.translate_inputs + args.translate_options)


def run_preprocess(args):
    logging.info("Running preprocessor.")
    logging.info("preprocessor arguments: %s" % args.preprocess_options)
    call_cmd(PREPROCESS, args.preprocess_options, stdin=args.preprocess_input)
    
def set_timeout(external_time_limit):
    if (external_time_limit is not None):
        elapsed_time = sum(os.times()[:4])        
        # Soft limit reached --> SIGXCPU.
        # Hard limit reached --> SIGKILL.
        hard_limit = int(math.floor(external_time_limit - elapsed_time))
        soft_limit = hard_limit - 1
        logging.info("remaining time: " + str(hard_limit))
        resource.setrlimit(resource.RLIMIT_CPU, (soft_limit, hard_limit))

def run_search(args):
    logging.info("Running search.")

    if args.debug:
        executable = os.path.join(SEARCH_DIR, "downward-debug")
    else:
        executable = os.path.join(SEARCH_DIR, "downward-release")
    logging.info("executable: %s" % executable)

    if args.portfolio:
        assert not args.search_options
        portfolio_runner.run(
            args.portfolio, executable, args.search_input, args.plan_file)
    else:
        if not args.search_options:
            raise SystemExit(
                "search needs --alias, --portfolio, or search options")
        if "--plan-file" not in args.search_options:
            args.search_options.extend(["--plan-file", args.plan_file])
        logging.info("final search options: %s" % args.search_options)
        
        
        if args.incremental_redblack:
            logging.info("performing incremental redblack search")

            EXIT_ONLY_RB_PLAN = 17  # FD exit code for EXIT_ONLY_RB_PLAN
            EXIT_UNSOLVED_INCOMPLETE = 5
            
            soft_time_limit, hard_time_limit = resource.getrlimit(resource.RLIMIT_CPU)            
            external_time_limit = None
            if soft_time_limit != resource.RLIM_INFINITY:
                external_time_limit = soft_time_limit
            elif hard_time_limit != resource.RLIM_INFINITY:
                external_time_limit = hard_time_limit
            
            return_code = EXIT_ONLY_RB_PLAN
            number_black = 0
            while (return_code == EXIT_ONLY_RB_PLAN):
                return_code = 0
                set_number_black_variables(args, number_black)
                try:
                    call_cmd(executable, args.search_options, stdin=args.search_input, preexec_funct=set_timeout(external_time_limit))
                except subprocess.CalledProcessError as err:
                    return_code = err.returncode
                    logging.info("exit code from redblack: " + str(return_code))
                    if (return_code != EXIT_ONLY_RB_PLAN):
                        print(err)
                        sys.exit(return_code)
                number_black += 1
        
        else:
            call_cmd(executable, args.search_options, stdin=args.search_input)
