#include "globals.h"

#include "option_parser.h"
#include "search_engine.h"
#include "timer.h"
#include "utilities.h"

#include "ext/tree_util.hh"
#include "painting.h"

#include <iostream>
#include <fstream>
#include <new>
#include <locale>
#include <string>

using namespace std;

vector<const Operator*> parse_planfile(string plan_filename);
string op_name_without_spaces(string op_name);


int main(int argc, const char **argv) {
    register_event_handlers();

    if (argc < 2) {
        cout << OptionParser::usage(argv[0]) << endl;
        exit_with(EXIT_INPUT_ERROR);
    }

    if (string(argv[1]).compare("--help") != 0){
	try {
	    OptionParser::parse_redblack_search(argc, argv);
	} catch (ArgError &error) {
	    cerr << error << endl;
	    OptionParser::usage(argv[0]);
	    exit_with(EXIT_INPUT_ERROR);
	}
        read_everything(cin);
    }

    SearchEngine *engine = 0;

    // The command line is parsed twice: once in dry-run mode, to
    // check for simple input errors, and then in normal mode.
    bool unit_cost = is_unit_cost();
    try {
        OptionParser::parse_cmd_line(argc, argv, true, unit_cost);
        engine = OptionParser::parse_cmd_line(argc, argv, false, unit_cost);
    } catch (ArgError &error) {
        cerr << error << endl;
        OptionParser::usage(argv[0]);
        exit_with(EXIT_INPUT_ERROR);
    } catch (ParseError &error) {
        cerr << error << endl;
        exit_with(EXIT_INPUT_ERROR);
    }

	if (g_rb_plan_sampling) {

		if (!g_use_redblack_search) {
			cout << "ERROR: The redblack search parameters need to be defined"
					<< endl;
			return 8;
		}
		if (!g_random_reorder) {
			cout
					<< "ERROR: need the number of samples with --random-reorder <NUM>"
					<< endl;
			return 8;
		}

		cout << "Initializing random sampling..." << endl;


		vector<const Operator *> plan = parse_planfile(g_input_plan_name);
		if (plan.empty())
			return 7;

		g_conflicts = countConflicts(plan);
		cout << "Conflicts of input plan: " << g_conflicts << endl;
		//print parsed plan
		//for (size_t i = 0; i < plan.size(); i++) {
		//	cout << plan[i]->get_name() << endl;
		//}

		cout << "Starting to sample reordered redblack plans..." << endl;
		cout << "Sampling start time: " << g_timer << endl;
		g_timer.resume();
		plan = sample_random_rb_plan(plan);
		g_timer.stop();

		save_plan(plan, 0);

		return 0;
	}

    Timer search_timer;
    engine->search();
    search_timer.stop();
    g_timer.stop();

    engine->save_plan_if_necessary();
    
    if (g_use_redblack_search && g_painting->is_incremental_search() && engine->found_solution() && !g_rb_plan_is_real_plan){
        cout << "time to find redblack plan: " << g_timer << endl;
        exit_with(EXIT_ONLY_RB_PLAN);
    }
    
    engine->statistics();
    engine->heuristic_statistics();
    cout << "Search time: " << search_timer << endl;
    cout << "Total time: " << g_timer << endl;

    if (engine->found_solution()) {
        exit_with(EXIT_PLAN_FOUND);
    } else {
        exit_with(EXIT_UNSOLVED_INCOMPLETE);
    }
}

vector<const Operator*> parse_planfile(string plan_filename){
	vector <const Operator*> plan;
	string line;
	ifstream plan_file;
	plan_file.open(plan_filename, ifstream::in);

	if(!plan_file.good()){
		cout << "opening file " << plan_filename  <<  "failed" << endl;
		return vector<const Operator * >();
	}

	vector <string> lines;
	while(getline(plan_file,line) && plan_file.good()){
		lines.push_back(line);
	}
	lines.pop_back();
	bool failed;
	for (size_t j = 0; j < lines.size(); j++) {
		failed = true;

		for (size_t i = 0; i < g_operators.size(); i++) {
			string op_name = op_name_without_spaces(g_operators[i].get_name());
			if (lines[j].find(op_name) != string::npos) {
				plan.push_back(&g_operators[i]);
				failed = false;
				break;
			}
		}
		if(failed){
			cout << "ERROR: parsing input_plan failed at line " << j+1 << endl;
			return vector<const Operator *>();
		}

	}
	plan_file.close();
	return plan;
}

string op_name_without_spaces(string op_name){
	locale loc;
	while (isspace(op_name[op_name.size() - 1], loc)) {
		op_name.erase(op_name.end() - 1);
	}
	return op_name;
}
