#include "successor_generator.h"

#include "globals.h"
#include "operator.h"
#include "painting.h"
#include "state.h"
#include "utilities.h"

#include <cstdlib>
#include <iostream>
#include <vector>
using namespace std;

class SuccessorGeneratorSwitch : public SuccessorGenerator {
    int switch_var;
    SuccessorGenerator *immediate_ops;
    vector<SuccessorGenerator *> generator_for_value;
    SuccessorGenerator *default_generator;
    
    SuccessorGeneratorSwitch(){};
public:
    SuccessorGeneratorSwitch(istream &in);
    
    virtual void generate_applicable_ops(const State &curr,
                                         vector<const Operator *> &ops);
    
    virtual void generate_applicable_ops(const PackedStateBin *buffer, 
                                         vector<const Operator *> &ops);
    
    virtual void _dump(string indent);
};

class SuccessorGeneratorGenerate : public SuccessorGenerator {
    friend class SuccessorGeneratorSwitch;
    
    vector<const Operator *> operators;
    
    SuccessorGeneratorGenerate(){};
public:
    SuccessorGeneratorGenerate(istream &in);
    
    virtual void generate_applicable_ops(const State &curr,
                                         vector<const Operator *> &ops);
    
    virtual void generate_applicable_ops(const PackedStateBin *buffer, 
                                         vector<const Operator *> &ops);
    
    virtual void _dump(string indent);
};

SuccessorGeneratorSwitch::SuccessorGeneratorSwitch(istream &in) {
    in >> switch_var;
    immediate_ops = read_successor_generator(in);
    for (int i = 0; i < g_variable_domain[switch_var]; ++i)
        generator_for_value.push_back(read_successor_generator(in));
    default_generator = read_successor_generator(in);
}

void SuccessorGeneratorSwitch::generate_applicable_ops(
    const State &curr, vector<const Operator *> &ops) {    
    immediate_ops->generate_applicable_ops(curr, ops);
    if (g_use_redblack_search && g_painting->is_red_var(switch_var)){
        for (int val = 0; val < g_variable_domain[switch_var]; ++val){
            if (curr.has_fact(switch_var, val)){
                generator_for_value[val]->generate_applicable_ops(curr, ops);
            }
        }
    } else {
        generator_for_value[curr[switch_var]]->generate_applicable_ops(curr, ops);
    }
    default_generator->generate_applicable_ops(curr, ops);
}

void SuccessorGeneratorSwitch::generate_applicable_ops(
    const PackedStateBin *buffer, vector<const Operator *> &ops) {    
    immediate_ops->generate_applicable_ops(buffer, ops);
    if (g_use_redblack_search && g_painting->is_red_var(switch_var)){
        for (int val = 0; val < g_variable_domain[switch_var]; ++val){
            if (g_state_packer->get_bit(buffer, switch_var, val)){
                generator_for_value[val]->generate_applicable_ops(buffer, ops);
            }
        }
    } else {
        generator_for_value[g_state_packer->get(buffer, switch_var)]->generate_applicable_ops(buffer, ops);
    }
    default_generator->generate_applicable_ops(buffer, ops);
}

void SuccessorGeneratorSwitch::_dump(string indent) {
    cout << indent << "switch on " << g_fact_names[switch_var][0] << endl;
    cout << indent << "immediately:" << endl;
    immediate_ops->_dump(indent + "  ");
    for (int i = 0; i < g_variable_domain[switch_var]; ++i) {
        cout << indent << "case " << g_fact_names[switch_var][i] << ":" << endl;
        generator_for_value[i]->_dump(indent + "  ");
    }
    cout << indent << "always:" << endl;
    default_generator->_dump(indent + "  ");
}

SuccessorGeneratorGenerate::SuccessorGeneratorGenerate(istream &in) {
    int count;
    in >> count;
    for (int i = 0; i < count; ++i) {
        int op_index;
        in >> op_index;
        operators.push_back(&g_operators[op_index]);
    }
}

void SuccessorGeneratorGenerate::generate_applicable_ops(const State &state,
                                                         vector<const Operator *> &ops) {
    if (g_use_redblack_search){
        ops.reserve(ops.size() + operators.size());
        for (auto const &op : operators){
            if (op->is_black_op()){
                bool changes_black = false;
                for (size_t eff = 0; eff < op->get_black_eff().size(); ++eff){
                    int var = op->get_black_eff()[eff]->var;
                    int val = op->get_black_eff()[eff]->val;
                    if (state[var] != val){
                        changes_black = true;
                        break;
                    }
                }
                if (changes_black){
                    ops.push_back(op);
                }
            }
        }
    } else {
        ops.insert(ops.end(), operators.begin(), operators.end());
    }
}

void SuccessorGeneratorGenerate::generate_applicable_ops(const PackedStateBin*,
                                                         vector<const Operator *> &ops) {
    ops.insert(ops.end(), operators.begin(), operators.end());
}

void SuccessorGeneratorGenerate::_dump(string indent) {
    for (size_t i = 0; i < operators.size(); ++i) {
        cout << indent;
        operators[i]->dump();
    }
}

SuccessorGenerator *read_successor_generator(istream &in) {
    string type;
    in >> type;
    if (type == "switch") {
        return new SuccessorGeneratorSwitch(in);
    } else if (type == "check") {
        return new SuccessorGeneratorGenerate(in);
    }
    cout << "Illegal successor generator statement!" << endl;
    cout << "Expected 'switch' or 'check', got '" << type << "'." << endl;
    exit_with(EXIT_INPUT_ERROR);
}
